/* Copyright (C) 2002 / 2003 AFE-GmdG */

/* Es _DARF_ nur GameLoopDemo.h includiert werden, weitere Header werden dort eingebunden. */
#include "GameLoopDemo.h"

/*
  Konstructor:
  In einem Konstructor sollten nur Membervariablen initialisiert werden. Ausfhrbarer Code
  zur Initialisierung gehrt in eine eigene Funktion, z.B. Init(). Dort steht dann auch
  schon ein gltiger this-Zeiger zur Verfgung - hier knnte er sich durchaus noch ndern.
*/
CGameLoopDemo::CGameLoopDemo(void) {
  m_hInstance=0; // Programminstanzvariable
  m_hWnd=0; // Handler des Hauptfensters

/*
  Die folgenden Mutexe sind zur Syncronisation zwischen den einzelnen Threads zustndig.
  In Zusammenarbeit mit Euch knnen wir vielleicht noch eine andere, vielleicht auch bessere
  Methode finden, die Threads ausreichend zu Syncronisieren.
*/
  m_hRunMutex=CreateMutex(NULL, FALSE, NULL); //Mutex fr die Windows-Nachrichtenschleife
  m_hRenderMutex=CreateMutex(NULL, FALSE, NULL); // Mutex fr die Renderfunktion
  m_hInputMutex=CreateMutex(NULL, FALSE, NULL); // Mutex fr die Input-Funktion

  m_lpInput=new CInput(); // Instanz der Input-Klasse anlegen
  m_ExitFail=FALSE; // Variable wird true, wenn ein Fehler auftritt
  m_ExitOK=FALSE; // Variable wird true, wenn Programm Fehlerlos beendet weden soll
}

/* Destructor */
CGameLoopDemo::~CGameLoopDemo(void) {
  m_hInstance=0;
  m_hWnd=0;
  CloseHandle(m_hRunMutex); // Mutexe mssen freigegeben werden, damit Windows sie wieder
  CloseHandle(m_hRenderMutex); // vergeben kann. Geschieht das nicht knnen in spteren
  CloseHandle(m_hInputMutex); // Programmaufrufen Deadlocks entstehen.
}

/*
  CreateMainWindow ist dafr zustndig, ein neues Hauptfenster zu erstellen.
*/
HWND CGameLoopDemo::CreateMainWindow() {
  WNDCLASSEX wc;
  wc.cbSize=sizeof(WNDCLASSEX);
  wc.style=CS_OWNDC;
  wc.lpfnWndProc=WndProc;
  wc.cbClsExtra=0;
  wc.cbWndExtra=0;
  wc.hInstance=m_hInstance;
  wc.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);
  wc.hCursor=LoadCursor(NULL, IDC_ARROW);
  wc.lpszMenuName=NULL;
  wc.lpszClassName=_T("Class_GameLoopDemo");
  wc.hIcon=LoadIcon(NULL, IDI_WINLOGO);
  wc.hIconSm=LoadIcon(NULL, IDI_WINLOGO);
  RegisterClassEx(&wc);
  return(CreateWindowEx(NULL, _T("Class_GameLoopDemo"), _T("Game Loop Demo"),
    WS_OVERLAPPEDWINDOW|WS_VISIBLE,
    CW_USEDEFAULT, CW_USEDEFAULT, 800, 600, NULL, NULL, m_hInstance, NULL));
}

/*
  Init wird nur einmal aufgerufen, und zwar von der WinMain-Funktion.
  Sollte sich herausstellen, dass init schon einmal aufgerufen wurde, gibt die
  Funktion false zurck. Kann das Hauptfenster aus irgend einem Grund nicht erstellt
  werden knnen, gibt die Funktion ebenfalls false zurck. Schlgt die
  Initialisierung der Input-Klasse fehl, wird auch false zurckgegeben.
  Im Platz fr weitere Initialisierungen kann z.B. DirectX initialisiert werden.
*/
BOOL CGameLoopDemo::Init(HINSTANCE hInstance) {
  if(!hInstance)
    return(FALSE);
  m_hInstance=hInstance;
  m_hWnd=CreateMainWindow();
  if(!m_hWnd)
    return(FALSE);
  if(!m_lpInput->Init(m_hInstance, m_hWnd, this))
    return(FALSE);
  /*
  Platz fr weitere Initialisierungen...
  */
  return(TRUE);
}

/*
  Run ist das Hauptstck von GameLoopDemo. Hier wird die Windowsnachrichtenschleife
  behandelt. Vorallem werden hier aber auch die neuen Threads erstellt
*/
BOOL CGameLoopDemo::Run() {
  MSG msg;
  DWORD ThreadID;
  /*
    Der RunMutex wird gesetzt. Kein anderer Thread kann den RunMutex jetzt reservieren bis
    er wieder freigegeben wird. INFINITE sagt aus, das unendlich lange gewartet werden soll,
    falls der Mutex bereits von einem Anderem Thread gesperrt worden ist.
  */
  WaitForSingleObject(m_hRunMutex, INFINITE);
  /*
    Die beiden Threads werden erstellt. Nach diesen Programmzeilen laufen die aufgerufenen
    Funktionen Unabhngig vom Hauptprogramm.
  */
  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) RenderProcedure, this, 0, &ThreadID);
  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) PollInputProcedure, m_lpInput, 0, &ThreadID);
  /*
    Der Timer ist dafr da, das in regelmigen Abstnden Messages die Nachrichtenschleife
    erreichen, da GetMessage Blockierend wirkt und so ein eventuell auftretendes
    m_ExitFail oder m_ExitOK nicht verarbeitet werden wrde.
  */
  SetTimer(m_hWnd, 1, 100, (TIMERPROC)NULL);
  while(GetMessage(&msg, NULL, 0, 0)&&!m_ExitFail&&!m_ExitOK) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  KillTimer(m_hWnd, 1); // Timer beenden - er wird nicht mehr gebraucht.
  ReleaseMutex(m_hRunMutex); // Mutex freigeben. Wartende Threads knnen ihn ketzt reservieren.
  /*
    Die beiden folgenden Zeilen warten, bis die entsprechenden Mutexe gesetzt werden knneen.
    Die Mutexe sind von den Entsprechenden Threads gesetzt und werden erst freigegeben, wenn
    die Threads beendet werden sollen. Damit erreiche ich das an dieser Stelle so lange gewartet
    wird, bis die Threads beendet worden sind.
  */
  WaitForSingleObject(m_hRenderMutex, INFINITE); // Warten, bis RenderMutex gesetzt werden kann.
  WaitForSingleObject(m_hInputMutex, INFINITE); // Warten bis InputMutex gesetzt werden kann.
  /*
    Rckgabewert ermitteln und setzen.
  */
  if(m_ExitFail)
    return(FALSE);
  return(TRUE);
}

/*
  Done ist fr einfache Aufrumarbeiten zustndig
*/
BOOL CGameLoopDemo::Done() {
  UnregisterClass(_T("Class_GameLoopDemo"), m_hInstance);
  if(m_lpInput) {
    m_lpInput->Release();
    m_lpInput=NULL;
  }
  return(TRUE);
}

/*
  Render ist die Funktion, in der grafische Ausgaben vorgenommen werden, bzw. andere Funktionen
  aufgerufen werden, die die grafischen Ausgaben bernehmen. Die Funktionalitt knnte in einem
  besser Strukturierten Programm auch in eine eigene Klasse verlagert werden, eine Aufgabe, die
  man sicherlich noch vornehmen knnte
*/
DWORD CGameLoopDemo::Render() {
  // Diese Funktion luft in einem eigenem Thread.
  /*
    RenderMutex setzen. Er wird erst beim Beenden der Funktion wieder freigegeben.
  */
  WaitForSingleObject(m_hRenderMutex, INFINITE);
  /*
    Diese Schleife wird solange aufgerufen, bis der RunMutex unter der Nachrichtenschleife in
    Run freigegeben wird. Solange der Mutex gesetzt ist, gibt es ein Timeout nach 10
    Millisekunden. Diese lassen dem Prozessor zeit, sich um andere Threads zu kmmern und
    sind ebenfalls dafr zustndig, die Prozessorlast nicht auf 100% zu steigern
  */
  while(WaitForSingleObject(m_hRunMutex, 10L)==WAIT_TIMEOUT) {
    // Hauptrenderschleife
    // hier nur Dinge einbauen, die etwas mit dem Rendern zu tun haben.
    // Ein GameStatus kann man hier auch unterbringen:
    /*
    switch(GameStatus) {
    case MainMenu:
      Klasse->RenderMainMenu();
      break;
    case MainGame:
      Klasse->RenderMainGame();
      break;
    case ExitGame:
      m_ExitOK=TRUE;
      break;
    }
    */
    // zum Beenden des Programmes von der Renderschleife aus
    // m_ExitOK=TRUE; // Programm beendet normal
    // m_ExitFail=TRUE; // Programm beendet mit Fehlercode -1
  }
  /*
    Der RenderMutex wird freigegeben, damit Run nach der Nachrichtenschleife weiss, das der
    Thread jetzt beendet ist.
  */
  ReleaseMutex(m_hRenderMutex);
  return(0);
}

HANDLE CGameLoopDemo::GetInputMutex() {
  return(m_hInputMutex);
}

HANDLE CGameLoopDemo::GetRunMutex() {
  return(m_hRunMutex);
}

/*
  WndProc berwacht die Windowsmessages und reagiert entsprechend. Da der Input von
  DirectInput und die Grafik von DirectDraw/3D verarbeitet werden sollte wird hier nur
  auf Systemnahe Meldungen reagiert - Wie z.B. das Beenden des Programmes
*/
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
  switch(msg) {
  case WM_DESTROY:
    PostQuitMessage(0);
    return(0);
    break;
  }
  return(DefWindowProc(hWnd, msg, wParam, lParam));
}

/*
  Da die Threadfunktionen statisch sein mssen, rufe ich einfach von diesen beiden
  die entsprechenden Funktionen der passenden Klassen auf.
*/
static DWORD WINAPI RenderProcedure(CGameLoopDemo *pInst) {
  return(pInst->Render());
}

static DWORD WINAPI PollInputProcedure(CInput *pInst) {
  return(pInst->PollInput());
}
